
#include <stdio.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <time.h>
#include <unistd.h>
#include <fcntl.h>
#include <signal.h>
#include <sys/stat.h>
#include <sys/select.h>
#include <stdlib.h>
#include <syslog.h>
#include <unistd.h>

#include <stdlib.h>
#include <process.h>
#include <sys/qnx_glob.h>

#define MAX_CONFSZ	255
#define SAT_MAXREAD	65535

//
// Default values.
//
#define SATLINK_DEV	"/dev/satlink"

#define CONFIG_FILE	"/usr/local/news/sat/etc/satcomd.conf"
#define ERROR_LOG	"/usr/local/news/sat/log/errors"
#define TRANS_LOG	"/usr/local/news/sat/log/transfer"

//
// Configuration file.
//
typedef struct ConfigRec {
	struct ConfigRec	*next;
	unsigned short		rulebase;
	int			active;
	char			*desc;
	char			*directory;
	char			*badprog;
	char			*goodprog;
} ConfigRec;

//
// This structure will hold the decoded file information
// from the downlink file header.
//
typedef struct FileHeader {
	char			fname[13];
	unsigned long		fsize;
	unsigned short		ms_time;
	unsigned short		ms_date;
	unsigned long		crc;
	unsigned short		descnum;
	unsigned short		decode;
	unsigned short		rbase;
	unsigned short		drop;
	unsigned short		misc;
	unsigned short		company;
	unsigned char		group;
	unsigned short		user;
	char			type;
	unsigned short		version;

	char			pathname[120];
	struct ConfigRec	*current;
	unsigned long		filecrc;
	int			outfd;
} FileHeader;

long tstart, tend;

//
// Define offset into the _56_ byte header for the
// above elements.
//
#define HDR_FNAME(a)	(a)
#define HDR_FSIZE(a)	(a+13)
#define HDR_TIME(a)	(a+17)
#define HDR_DATE(a)	(a+19)
#define HDR_SUM(a)	(a+21)
#define HDR_CRC(a)	(a+23)
#define HDR_DESCTAB(a)	(a+27)
#define HDR_DECODE(a)	(a+29)
#define HDR_RULEBASE(a)	(a+31)
#define HDR_DROP(a)	(a+33)
#define HDR_MISC(a)	(a+35)
#define HDR_YAGI(a)	(a+37)
#define HDR_COMPANY(a)	(a+42)
#define HDR_GROUP(a)	(a+44)
#define HDR_USER(a)	(a+45)
#define HDR_TYPE(a)	(a+47)
#define HDR_VERSION(a)	(a+48)
#define HDR_FILLER(a)	(a+50)

//
//  Application Global variables.
//
ConfigRec	*config	= (ConfigRec *) NULL;
FileHeader	header;

//
// File descriptor for the satellite device driver.
//
int		satlink;

//
// Application configuration defaults.
//
char		*device_name	= SATLINK_DEV;
char		*config_file	= CONFIG_FILE;
int		debug		= 0;
int		visual		= 0;

//
// CCITT CRC 32
//

unsigned long crc_32_tab[256] = {
0x00000000, 0x77073096, 0xee0e612c, 0x990951ba, 0x076dc419, 0x706af48f, 0xe963a535, 0x9e6495a3,
0x0edb8832, 0x79dcb8a4, 0xe0d5e91e, 0x97d2d988, 0x09b64c2b, 0x7eb17cbd, 0xe7b82d07, 0x90bf1d91,
0x1db71064, 0x6ab020f2, 0xf3b97148, 0x84be41de, 0x1adad47d, 0x6ddde4eb, 0xf4d4b551, 0x83d385c7,
0x136c9856, 0x646ba8c0, 0xfd62f97a, 0x8a65c9ec, 0x14015c4f, 0x63066cd9, 0xfa0f3d63, 0x8d080df5,
0x3b6e20c8, 0x4c69105e, 0xd56041e4, 0xa2677172, 0x3c03e4d1, 0x4b04d447, 0xd20d85fd, 0xa50ab56b,
0x35b5a8fa, 0x42b2986c, 0xdbbbc9d6, 0xacbcf940, 0x32d86ce3, 0x45df5c75, 0xdcd60dcf, 0xabd13d59,
0x26d930ac, 0x51de003a, 0xc8d75180, 0xbfd06116, 0x21b4f4b5, 0x56b3c423, 0xcfba9599, 0xb8bda50f,
0x2802b89e, 0x5f058808, 0xc60cd9b2, 0xb10be924, 0x2f6f7c87, 0x58684c11, 0xc1611dab, 0xb6662d3d,
0x76dc4190, 0x01db7106, 0x98d220bc, 0xefd5102a, 0x71b18589, 0x06b6b51f, 0x9fbfe4a5, 0xe8b8d433,
0x7807c9a2, 0x0f00f934, 0x9609a88e, 0xe10e9818, 0x7f6a0dbb, 0x086d3d2d, 0x91646c97, 0xe6635c01,
0x6b6b51f4, 0x1c6c6162, 0x856530d8, 0xf262004e, 0x6c0695ed, 0x1b01a57b, 0x8208f4c1, 0xf50fc457,
0x65b0d9c6, 0x12b7e950, 0x8bbeb8ea, 0xfcb9887c, 0x62dd1ddf, 0x15da2d49, 0x8cd37cf3, 0xfbd44c65,
0x4db26158, 0x3ab551ce, 0xa3bc0074, 0xd4bb30e2, 0x4adfa541, 0x3dd895d7, 0xa4d1c46d, 0xd3d6f4fb,
0x4369e96a, 0x346ed9fc, 0xad678846, 0xda60b8d0, 0x44042d73, 0x33031de5, 0xaa0a4c5f, 0xdd0d7cc9,
0x5005713c, 0x270241aa, 0xbe0b1010, 0xc90c2086, 0x5768b525, 0x206f85b3, 0xb966d409, 0xce61e49f,
0x5edef90e, 0x29d9c998, 0xb0d09822, 0xc7d7a8b4, 0x59b33d17, 0x2eb40d81, 0xb7bd5c3b, 0xc0ba6cad,
0xedb88320, 0x9abfb3b6, 0x03b6e20c, 0x74b1d29a, 0xead54739, 0x9dd277af, 0x04db2615, 0x73dc1683,
0xe3630b12, 0x94643b84, 0x0d6d6a3e, 0x7a6a5aa8, 0xe40ecf0b, 0x9309ff9d, 0x0a00ae27, 0x7d079eb1,
0xf00f9344, 0x8708a3d2, 0x1e01f268, 0x6906c2fe, 0xf762575d, 0x806567cb, 0x196c3671, 0x6e6b06e7,
0xfed41b76, 0x89d32be0, 0x10da7a5a, 0x67dd4acc, 0xf9b9df6f, 0x8ebeeff9, 0x17b7be43, 0x60b08ed5,
0xd6d6a3e8, 0xa1d1937e, 0x38d8c2c4, 0x4fdff252, 0xd1bb67f1, 0xa6bc5767, 0x3fb506dd, 0x48b2364b,
0xd80d2bda, 0xaf0a1b4c, 0x36034af6, 0x41047a60, 0xdf60efc3, 0xa867df55, 0x316e8eef, 0x4669be79,
0xcb61b38c, 0xbc66831a, 0x256fd2a0, 0x5268e236, 0xcc0c7795, 0xbb0b4703, 0x220216b9, 0x5505262f,
0xc5ba3bbe, 0xb2bd0b28, 0x2bb45a92, 0x5cb36a04, 0xc2d7ffa7, 0xb5d0cf31, 0x2cd99e8b, 0x5bdeae1d,
0x9b64c2b0, 0xec63f226, 0x756aa39c, 0x026d930a, 0x9c0906a9, 0xeb0e363f, 0x72076785, 0x05005713,
0x95bf4a82, 0xe2b87a14, 0x7bb12bae, 0x0cb61b38, 0x92d28e9b, 0xe5d5be0d, 0x7cdcefb7, 0x0bdbdf21,
0x86d3d2d4, 0xf1d4e242, 0x68ddb3f8, 0x1fda836e, 0x81be16cd, 0xf6b9265b, 0x6fb077e1, 0x18b74777,
0x88085ae6, 0xff0f6a70, 0x66063bca, 0x11010b5c, 0x8f659eff, 0xf862ae69, 0x616bffd3, 0x166ccf45,
0xa00ae278, 0xd70dd2ee, 0x4e048354, 0x3903b3c2, 0xa7672661, 0xd06016f7, 0x4969474d, 0x3e6e77db,
0xaed16a4a, 0xd9d65adc, 0x40df0b66, 0x37d83bf0, 0xa9bcae53, 0xdebb9ec5, 0x47b2cf7f, 0x30b5ffe9,
0xbdbdf21c, 0xcabac28a, 0x53b39330, 0x24b4a3a6, 0xbad03605, 0xcdd70693, 0x54de5729, 0x23d967bf,
0xb3667a2e, 0xc4614ab8, 0x5d681b02, 0x2a6f2b94, 0xb40bbe37, 0xc30c8ea1, 0x5a05df1b, 0x2d02ef8d
};

unsigned long UpdCrc32 (unsigned long crc, unsigned char octet)
{
    return (crc_32_tab[ (unsigned char) (crc ^ (unsigned long) octet)] ^ ((crc>>8) & 0x00ffffff));
}

//
// This function will open the satellite device driver. If
// any errors it will log them, then return a false status
// to the caller.
//
int
InitSatComm(char *devname)
{
	if((satlink = open(devname,O_RDONLY)) < 0) {
		//
		// Error opening the satlink device.
		//
			printf("satcomd: Failed to open the satellite device. (%s)\n", devname );
		return(0);
	}
	if ( visual )
		printf("satcomd: Open satellite device.\n");
	return(1);
}

//
// This function do any closing of the satellite device driver.
//
void
ShutSatComm()
{
	if ( visual )
		printf("satcomd: Closing satellite device.\n");
	close(satlink);
}

//
// This function gets called at the beginning of a file transfer.
//
static int
StartFile()
{
	ConfigRec	*tmpconf;

	//
	// See if we have a configuration record to match the incoming
	// file. If we don't then just ignore the incoming file.
	//
	tmpconf = config;
	while(tmpconf) {
		if(tmpconf->rulebase == header.rbase)
			break;
		tmpconf = tmpconf->next;
	}

	//
	// Set the current configuration information to the tmpconf
	// pointer, which could be null, in the case of an unknown file
	// type or it could be the config record for the given file type.
	//
	header.current = tmpconf;

	//
	// Log the fact that we are receiving a file.
	//
	if ( visual )
		printf("satcomd: Incoming satellite transmittion, %s is %ld bytes long.\n",
			header.fname, header.fsize );
	
	//
	// Ok, now depending on configuration, we'll open a file to put this
	// data into.
	//
	if(header.current) {
		//
		// we know about the file, see if we're allowed to receive it
		// or not.
		//
		if(header.current->active) {
			//
			// Build this final pathname from the configuration
			// information.
			//
			sprintf(header.pathname,"%s/%s",header.current->directory,header.fname);

			//
			// now alert the user.
			//
			//if(debug)
				//syslog(LOG_INFO,"Creating file %s",header.pathname);

			//
			// Then create the file.
			//
			header.outfd = creat(header.pathname,S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP|S_IROTH|S_IWOTH);
			if(header.outfd < 0) {
				//
				// error occurred while opening incoming file.
				//
				if ( visual )
					printf( "satcomd: Error %d creating file %s\n",errno,header.pathname);
				return(0);
			}
		} else {
			//
			// We're not dealing this file type. Log this fact to
			// the system log.
			//
			header.outfd = -1;

			if ( visual )
				printf( "satcomd: Reception of this file type %d, has been disabled.\n",header.rbase);
		}
	} else {
		//
		// The file being transmitted is not a file type that
		// we're geared to deal with, so just ignore it, but log
		// the fact to the syslog.
		//
		if ( visual )
			printf("satcomd: File type %d unknown, ignoring transmittion.\n",header.rbase);
	}

	tstart = time(NULL);

	//
	// All's well that ends well
	//
	return(1);
}

static int
EndFile()
{
	char	*program;
	pid_t	pid;
	FILE	*fp;

	tend = time(NULL);

	if(header.current) {
		if(header.current->active) {
			//
			// first things first, close the incoming file handle
			// to flush the data.
			//
			close(header.outfd);

			if(header.crc != header.filecrc) {
				//
				// the computed CRC doesn't match the declared
				// CRC in the file header.
				//
				if ( visual )
					printf("satcomd: File %s was received with an invalid CRC\n",header.fname);

				fp = fopen( ERROR_LOG, "a" );
				if ( fp )
					{
					long clk = time(NULL);
					struct tm *st = localtime( &clk );
					fprintf( fp, "%02d/%02d/%02d %02d:%02d:%02d %8ld %s (CRC Hdr %08lx File %08lx)\n",
						st->tm_mday, st->tm_mon+1, st->tm_year,
						st->tm_hour, st->tm_min, st->tm_sec,
						header.fsize, header.fname,	header.crc, header.filecrc );
					fclose( fp );
					}

				program = header.current->badprog;
			} else {
				//
				//  File %s was received correctly, header.fname 
				//
				program = header.current->goodprog;

				fp = fopen( TRANS_LOG, "a" );
				if ( fp )
					{
					long clk = time(NULL);
					struct tm *st = localtime( &clk );
					fprintf( fp, "%02d/%02d/%02d %02d:%02d:%02d %8ld %-11.11s %-3ld\n",
						st->tm_mday, st->tm_mon+1, st->tm_year,
						st->tm_hour, st->tm_min, st->tm_sec,
						header.fsize, header.fname, (tend-tstart) );
					fclose( fp );
					}

				qnx_spawn_options.flags |= _SPAWN_BGROUND|_SPAWN_NOZOMBIE;
				qnx_spawn_options.priority = 10;
				spawnl( 1, "/usr/local/bin/satxb", "satxb", header.fname, NULL );
			}
		}
	}

	return(1);
}

void
HeaderHunt()
{
	unsigned long	SOH;

	if ( visual )
		printf("satcomd: Looping to look for a Start of Header signature.\n");

	//
	// Loop until we see the start of header signature.
	// we will always be successful with this, because
	// we will not return until we have found it.
	//
	while(SOH != 0x0000FFFFL)
		read(satlink,&SOH,sizeof(SOH));
}

int
HeaderRead()
{
	unsigned char	hbuffer[56];
	int		remain;
	int		len;

	//
	// Read the header data in
	//
	remain = sizeof(hbuffer);
	while(remain) {
		len	= read(satlink,hbuffer+(sizeof(hbuffer) - remain),remain);
		if(len < 0) {
			return(0);
		}

		remain -= len;
	}

	//
	// Now decode the header buffer placing this information into
	// the header data structure. If this function returns a false
	// then the header was invalid.
	//

	//
	// Clear out the header, so that we don't have any
	// question as to the validity of this data structure.
	//
	bzero(&header,sizeof(struct FileHeader));

	//
	// Check to see if this header is a
	// valid header.
	//
	if(strcmp(HDR_YAGI(hbuffer),"YAGI")) {
		if ( visual )
			printf("satcomd: Invalid header signature.\n");

		return(0);
	}
	
	memcpy(&header.company,	HDR_COMPANY(hbuffer),	sizeof(unsigned short));
	memcpy(&header.group,	HDR_GROUP(hbuffer),	sizeof(unsigned char));
	memcpy(&header.user,	HDR_USER(hbuffer),	sizeof(unsigned short));

	if((header.company != 0xFFFF) && ((header.group) != 0xFF) && ((header.user) != 0xFFFF)) {
		if ( visual )
			printf("satcomd: Company/Group/User is not broadcast!\n");

		return(0);
	}

	//
	// If we get this far it means that we can decode
	// the header.
	//
	memcpy(&header.fname,	HDR_FNAME(hbuffer),	13);
	memcpy(&header.ms_time,	HDR_TIME(hbuffer),	sizeof(unsigned short));
	memcpy(&header.ms_date,	HDR_DATE(hbuffer),	sizeof(unsigned short));
	memcpy(&header.crc,	HDR_CRC(hbuffer),	sizeof(unsigned long));

	memcpy(&header.descnum,	HDR_DESCTAB(hbuffer),	sizeof(unsigned short));
	memcpy(&header.decode,	HDR_DECODE(hbuffer),	sizeof(unsigned short));
	memcpy(&header.rbase,	HDR_RULEBASE(hbuffer),	sizeof(unsigned short));
	memcpy(&header.drop,	HDR_DROP(hbuffer),	sizeof(unsigned short));
	memcpy(&header.misc,	HDR_MISC(hbuffer),	sizeof(unsigned short));
	memcpy(&header.type,	HDR_TYPE(hbuffer),	sizeof(unsigned short));
	memcpy(&header.version,	HDR_VERSION(hbuffer),	sizeof(unsigned short));

	memcpy(&header.fsize,	HDR_FSIZE(hbuffer),	sizeof(unsigned long));

	//
	// Return a true value stating that we were successful.
	//
	return(1);
}

int
FileRead()
{
	unsigned char	fbuffer[SAT_MAXREAD];
	unsigned long	fptr;
	unsigned long	len;
	unsigned char	*byte;
	unsigned long	readsiz;
	fd_set		read_set;
	struct timeval	timeout;
	int		retval;

	header.filecrc	= 0xFFFFFFFFL;
	fptr	= 0;

	//
	// Go into a loop and read the data from the satellite driver as it
	// comes in, place it into the file that was openned in the
	// StartFile() function.
	//
	while(fptr < header.fsize) {

		//
		// Setup for the select statement.
		//
		FD_ZERO(&read_set);
		FD_SET(satlink,&read_set);
		timeout.tv_sec = 1;
		timeout.tv_usec = 0;

		if((retval = select(satlink+1,&read_set,0,0,&timeout)) < 0) {
			//
			// Wow! The select had failed! Report this
			// at once to my loyal subjects!
			//
			printf("satcomd: Error during the select statement waiting for data.\n");
			return(0);
		}

		if(retval == 0) {
			//
			// The data stream has stopped during a file download.
			// that means that we over shot the mark. Reset! Get
			// ready for the next file!
			//
			return(1);
		}

		readsiz = ((header.fsize - fptr) > SAT_MAXREAD) ? SAT_MAXREAD : (header.fsize - fptr);

		len = read(satlink,fbuffer,readsiz);
		if(len <= 0) {
			//
			// Houston - We have a problem.
			//
			//syslog(LOG_ERR,"Error reading from the satellite device driver.");
			return(0);
		}
		fptr += len;

		//
		// Compute the CRC on the fly.
		//
		for(byte=fbuffer;byte < (fbuffer + len);byte++)
			header.filecrc = UpdCrc32(header.filecrc,*byte);

		//
		// Write the data out to the working file.
		//
		if(header.current) {
			if(header.current->active) {
				if(write(header.outfd,fbuffer,len) != len) {
					//
					// For some reason we we're unable to write
					// data out to the file. return a false status
					// to the caller.
					//
					//syslog(LOG_ERR,"Error writing received satellite data to %s",header.fname);
					return(0);
				}
			}
		}
	}

	//
	// Invert the bits in the CRC. So that they will match up with what
	// is in the header.
	//
	header.filecrc = ~header.filecrc;

	//
	// All's well that ends well.
	//
	return(1);
}

void
DoReceive()
{
	//
	// Open and Initialize the satellite device driver.
	//
	InitSatComm(device_name);

	while(1) {
		//
		// Go into a loop looking for the header signature of a
		// file. If we get it, proceed to read the file into
		// memory. This function will not return until we have
		// a valid Start of Header signature.
		//
		HeaderHunt();

		//
		// Now that we have the start of header signature
		// proceed to read the header and process the incoming
		// file. If this function returns a false value, then
		// the header didn't pass the sanity check restart this
		// while loop to look for another SOH signature.
		//
		if(!HeaderRead())
			continue;

		//
		// The header was good, now we have to call the start of
		// file function to process any incoming files, for example
		// we'll check to see if we want the file, if so, create the
		// file in the correct working directory.
		//
		if(!StartFile()) {
			//
			// An error occurred while openning the file and
			// starting the transfer. We'll need to log this
			// to the syslog and then restart the while loop.
			//
			continue;
		}

		if(!FileRead()) {
			//
			// An error occurred while downloading the file.
			// log this to the syslog, zero the current filecrc
			// and allow the EndFile() function to clean up the
			// mess and run the bad file program.
			//
			header.filecrc = 0;
		}

		//
		// Process any end of file things, like verifing CRC
		// of the file and running any auto processing programs
		// against the received file.
		//
		EndFile();

		//
		// Go around and start the next file.
		//
	}

	//
	// Close down the satellite driver.
	//
	ShutSatComm();
}

int
ReadConfigFile(char *cfile)
{
	ConfigRec	*celement;
	ConfigRec	*tmprec,*tmprec2;
	FILE		*fd;
	char		line[MAX_CONFSZ];
	char		*field;

	//
	// Temporary holding variables.
	//
	unsigned short	rulebase;
	int		active;
	char		*desc;
	char		*directory;
	char		*badprog;
	char		*goodprog;

	//
	// Check to see if we've done this before
	// if so clear out the old config information.
	//

	if(config) {
		//if(debug>1)
			//syslog(LOG_INFO,"Clearing old config records out.\n");

		tmprec = config;
		while(tmprec) {
			if(tmprec->desc)
				free(tmprec->desc);
			if(tmprec->directory)
				free(tmprec->directory);
			if(tmprec->badprog)
				free(tmprec->badprog);
			if(tmprec->goodprog)
				free(tmprec->goodprog);
			tmprec2	= tmprec;
			tmprec	= tmprec->next;
			free(tmprec2);
		}
		config = (ConfigRec *) NULL;
	}

	//
	// Open the specified configuration file.
	//
	fd = fopen(cfile,"r");
	if(fd == (FILE *) NULL) {
		//
		// Ouch, cann't open the config file. Scream and die.
		//
		//syslog(LOG_ERR,"Failed to open configuration file %s",cfile);
		return(0);
	}

	//
	// Read the config file in and process it.
	//
	while(!feof(fd)) {
		if(fgets(line,MAX_CONFSZ,fd) != (char *) NULL) {
			if((line[0] == '#') || (!strlen(line)))
				continue;

			field = strtok(line,":\r\n");
			if(field != (char *) NULL) {
				rulebase = atoi(field);
			} else continue;

			field = strtok((char *) NULL,":\r\n");
			if(field != (char *) NULL) {
				active = atoi(field);
			} else continue;

			field = strtok((char *) NULL,":\r\n");
			if(field != (char *) NULL) {
				desc = strdup(field);
			} else continue;

			field = strtok((char *) NULL,":\r\n");
			if(field != (char *) NULL) {
				directory = strdup(field);
			} else continue;

			field = strtok((char *) NULL,":\r\n");
			if(field != (char *) NULL) {
				badprog = strdup(field);
			} else continue;

			field = strtok((char *) NULL,":\r\n");
			if(field != (char *) NULL) {
				goodprog = strdup(field);
			} else continue;

			celement = (ConfigRec *) malloc(sizeof(struct ConfigRec));
			if(celement == (ConfigRec *) NULL) {
				//
				// We ran out of memory for the ConfigRec
				// entry, this is a fatal error. Scream and
				// die a very bad death.
				//
				//syslog(LOG_ERR,"Unable to allocate memory for Config Record.");
				return(0);
			}

			celement->rulebase = rulebase;
			celement->active = active;
			celement->desc = desc;
			celement->directory = directory;
			celement->badprog = badprog;
			celement->goodprog = goodprog;

			celement->next = config;
			config = celement;
		}
	}

	fclose(fd);
	return(1);
}

int main(int argc,char *argv[])
{
	char	optc;
	int		i;

	//
	// Process the command line arguments.
	//
	while(1) {
		optc = getopt(argc,argv,"d:f:qvV");
		if(optc == -1)
			break;

		switch(optc) {
			case 'd':
				device_name = strdup(optarg);
				break;
			case 'f':
				config_file = strdup(optarg);
				break;
			case 'v':
				debug++;
				break;
			case 'q':
				visual++;
				break;
			case 'V':
				printf("Satcomd Version 1.1\n");
				break;
			default:
		}
	}

	if ( !visual )
		for( i = _SIGMIN; i <= _SIGMAX; ++i)
			if( i != SIGTERM )
				signal( i, SIG_IGN );

	if(!ReadConfigFile(config_file)) {
		//
		// This means that there was an error while reading the
		// configuration file. We'll need to log it to the syslog
		// and exit the program with an error.
		//
		exit(1);
	}

	//
	// Do not loose data !
	//
	nice( 12 );

	//
	// Begin the process of receiving the files.
	//
	DoReceive();

	return(0);
}

